import json
import os
import random
import re
import sys
import tempfile
import time
import unicodedata
from urllib.parse import urljoin
import uuid
import requests
import speech_recognition
from pydub import AudioSegment

# web_scraper.py
import requests
from bs4 import BeautifulSoup
from termcolor import cprint

from utils.app_config import USER_AGENTS

try:
    from selenium import webdriver
    from selenium.webdriver.common.by import By

    from selenium.webdriver.firefox.service import Service as FireFoxService
    from selenium.webdriver.firefox.options import Options as FireFoxOptions

    from selenium.webdriver.chrome.service import Service
    from selenium.webdriver.chrome.options import Options

    from selenium.webdriver.support.ui import WebDriverWait
    from selenium.webdriver.common.action_chains import ActionChains
    from selenium.webdriver.support import expected_conditions as EC
    from selenium.webdriver.common.by import By
    from selenium.webdriver.support import expected_conditions as ec
    from selenium.webdriver.remote.webelement import WebElement
    from selenium.webdriver.chrome.webdriver import WebDriver
    from selenium.webdriver.support.ui import WebDriverWait
    from selenium.common.exceptions import TimeoutException
except ImportError:
    cprint(
        "Selenium is not installed. Please install it using 'pip install selenium'.",
        "red",
        file=sys.stderr,
    )
    sys.exit(1)


class SearchResult:
    def __init__(self, url, title, description):
        self.url = url
        self.title = title
        self.description = description

    def __repr__(self):
        return f"SearchResult(url={self.url}, title={self.title}, description={self.description})"


def you_are_human(response):
    """_summary_

    Args:
        response (_type_): _description_

    Returns:
        _type_: _description_
    """
    soup = BeautifulSoup(response, "html.parser")
    if soup.find("div", {"class": "g-recaptcha"}):
        # https://www.zenrows.com/blog/please-verify-you-are-human#headless-browser
        return True
    return False


# https://github.com/thicccat688/selenium-recaptcha-solver/blob/main/selenium_recaptcha_solver/solver.py#L153
def by_pass_captcha(driver):
    """_summary_

    Args:
        driver (_type_): _description_

    Raises:
        Exception: _description_
        Exception: _description_
    """
    try:
        WebDriverWait(driver, 10).until(
            EC.frame_to_be_available_and_switch_to_it(
                (
                    By.CSS_SELECTOR,
                    "iframe[name^='a-'][src^='https://www.google.com/recaptcha/api2/anchor?']",
                )
            )
        )
        time.sleep(random.uniform(2, 10))
        cprint("Click reCAPTCHA ...", color="yellow", file=sys.stderr)
        WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, "//span[@id='recaptcha-anchor']"))
        ).click()
        time.sleep(random.uniform(2, 10))
    except TimeoutException:
        cprint("reCAPTCHA not found ...", color="red", file=sys.stderr)
        return

    # If the captcha image audio is available, locate it. Otherwise, skip to the next line of code.
    try:
        cprint(
            "Bypassing reCAPTCHA - Audio phase - captcha image audio is available, locate it.",
            color="yellow",
            file=sys.stderr,
        )
        WebDriverWait(driver, 20).until(
            EC.element_to_be_clickable((By.XPATH, '//*[@id="recaptcha-audio-button"]'))
        ).click()
        time.sleep(random.uniform(2, 10))
    except TimeoutException:
        cprint(
            "Bypassing reCAPTCHA - Audio phase - captcha image audio is not available, skip to the next line of code.",
            color="yellow",
            file=sys.stderr,
        )
        pass

    try:
        bypass_captcha_audio_phase(driver)
        time.sleep(random.uniform(2, 10))
    except Exception as e:
        cprint(
            f"Error bypassing reCAPTCHA - Audio phase: {e}",
            color="red",
            file=sys.stderr,
        )
        return

    # Locate verify button and click it via JavaScript
    WebDriverWait(driver, 20).until(
        EC.element_to_be_clickable((By.ID, "recaptcha-verify-button"))
    ).click()
    time.sleep(random.uniform(2, 10))
    try:
        cprint(
            "Bypassing reCAPTCHA - Audio phase solve more",
            color="yellow",
            file=sys.stderr,
        )

        WebDriverWait(driver, 20).until(
            EC.visibility_of_element_located(
                (
                    By.XPATH,
                    '//div[normalize-space()="Multiple correct solutions required - please solve more."]',
                )
            )
        )
        time.sleep(random.uniform(2, 10))
        bypass_captcha_audio_phase(driver)
        time.sleep(random.uniform(2, 10))
        # Locate verify button again to avoid stale element reference and click it via JavaScript
        WebDriverWait(driver, 20).until(
            EC.element_to_be_clickable((By.ID, "recaptcha-verify-button"))
        ).click()
    except TimeoutException:
        cprint(
            "Bypassing reCAPTCHA - Audio phase - Multiple correct solutions required - please solve more.",
            color="yellow",
            file=sys.stderr,
        )
        pass

def slugify(value, allow_unicode=False):
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value.lower())
    return re.sub(r'[-\s]+', '-', value).strip('-_')

def parse_headers(headers):
    headers = headers.replace('\\n', '\n')
    sorted_headers = {}
    matches = re.findall(r'(.*):\s(.*)', headers)
    for match in matches:
        header = match[0]
        value = match[1]
        try:
            if value[-1] == ',':
                value = value[:-1]
            sorted_headers[header] = value
        except IndexError:
            pass
    return sorted_headers

def bypass_captcha_audio_phase(driver):
    recognizer = speech_recognition.Recognizer()

    try:
        cprint(
            "Bypassing reCAPTCHA - Audio phase download",
            color="yellow",
            file=sys.stderr,
        )
        # Locate audio challenge download link
        # Get the audio challenge URI from the download link
        download_link = driver.find_element(
            By.XPATH, "rc-audiochallenge-tdownload-link"
        )
        cprint("Download link found", color="green", file=sys.stderr)

    except TimeoutException:
        raise Exception("Google has detected automated queries. Try again later.")
    except Exception as e:
        raise Exception(f"Error locating download link: {e}")

    # Create temporary directory and temporary files
    tmp_dir = tempfile.gettempdir()

    id_ = uuid.uuid4().hex

    mp3_file, wav_file = os.path.join(tmp_dir, f"{id_}_tmp.mp3"), os.path.join(
        tmp_dir, f"{id_}_tmp.wav"
    )
    cprint(
        f"Temporary files created: {mp3_file}, {wav_file}",
        color="yellow",
        file=sys.stderr,
    )
    tmp_files = {mp3_file, wav_file}

    with open(mp3_file, "wb") as f:
        link = download_link.get_attribute("href")

        audio_download = requests.get(url=link, allow_redirects=True)

        f.write(audio_download.content)

        f.close()

    # Convert MP3 to WAV format for compatibility with speech recognizer APIs
    AudioSegment.from_mp3(mp3_file).export(wav_file, format="wav")

    # Disable dynamic energy threshold to avoid failed reCAPTCHA audio transcription due to static noise
    recognizer.dynamic_energy_threshold = False

    with speech_recognition.AudioFile(wav_file) as source:
        audio = recognizer.listen(source)

        try:
            recognized_text = recognizer.recognize_google(audio, "en-US")

        except speech_recognition.UnknownValueError:
            raise Exception(
                "Speech recognition API could not understand audio, try again"
            )

    # Clean up all temporary files
    for path in tmp_files:
        if os.path.exists(path):
            os.remove(path)

    # Write transcribed text to iframe's input box
    response_textbox = driver.find_element(By.ID, "audio-response")

    for c in recognized_text:
        response_textbox.send_keys(c)
        time.sleep(random.uniform(0.05, 0.1))

    # # Switch to iframe that directly houses pX "Press & Hold" button
    # WebDriverWait(driver, timeout=300).until(
    #     EC.frame_to_be_available_and_switch_to_it("iframe_name_or_id")
    # )
    # # Get button element
    # btn = driver.find_element(By.XPATH, "//xpath_to_button")
    # # Initialize for low-level interactions
    # action = ActionChains(driver)
    # action.click_and_hold(btn)
    # # Initiate clich and hold action on button
    # action.perform()
    # # Keep holding for 10s
    # time.sleep(10)
    # # Release button
    # action.release(btn)


def parse_google_search_results(proxies, advanced, full_query, response):
    """_summary_

    Args:
        proxies (_type_): _description_
        advanced (_type_): _description_
        full_query (_type_): _description_
        response (_type_): _description_

    Returns:
        _type_: _description_
    """
    urls = []
    soup = BeautifulSoup(response, "html.parser")
    result_block = soup.find_all("div", attrs={"class": "g"})
    cprint(
        f"Potentially {len(result_block)} links ...",
        "yellow",
        file=sys.stderr,
    )
    if len(result_block) == 0:
        cprint(
            f"No results found for parsing of {full_query} with proxy {proxies}\nTrying new parsing method",
            "yellow",
            file=sys.stderr,
        )
        try:
            filename = "outputs/html_google_todo/" + slugify(f"google_search_{full_query}") + ".html"
            with open(
                filename, "w"
            ) as f:
                f.write(response)
        except Exception as e:
            cprint(
                f"Error writing to file: {e}",
                "red",
                file=sys.stderr,
            )
        # Locate all <a> tags that contain the search results
        for a_tag in soup.find_all("a", href=True):
            # Extract the href attribute
            href = a_tag["href"]
            # Only consider hrefs that start with '/url?'
            if href.startswith("/url?"):
                # Extract the actual URL using regex
                url_match = re.search(r"(https?://[^&]+)", href)
                if url_match:
                    url = url_match.group(0)
                    cprint(f"Checking for url: {url}", color="yellow", file=sys.stderr)
                    # Extract the title (text within <div> with specific class)
                    title_tag = a_tag.find("h3") or a_tag.find(
                        "div", class_="BNeawe vvjwJb AP7Wnd UwRFLe"
                    )
                    title = title_tag.get_text() if title_tag else None
                    if title:
                        cprint(
                            f"Link appended to potential urls: {url}",
                            "green",
                            file=sys.stderr,
                        )
                        urls.append(url)
            else:
                pass
    else:
        for result in result_block:
            # Find link, title, description
            link = result.find("a", href=True)
            title = result.find("h3")
            description_box = result.find("div", {"style": "-webkit-line-clamp:2"})
            if description_box:
                description = description_box.text
                if link and title and description:
                    cprint(
                        f"Link appended to potential urls: {link['href']}",
                        "green",
                        file=sys.stderr,
                    )
                    if advanced:
                        urls.append(SearchResult(link["href"], title.text, description))
                    else:
                        urls.append(link["href"])
    cprint(
        f"Done parsing search results - {len(urls)} potential new links",
        "green",
        file=sys.stderr,
    )
    return urls


def fetch_html_content(url, config, proxy=None, headers=None):
    """
    Fetch the HTML content of a given URL.

    Args:
        url (str): The URL to fetch the content from.

    Returns:
        str: HTML content of the page or None if request fails.
    """
    try:
        cprint(f"Fetching {url}", color="yellow", file=sys.stderr)
        if proxy and "username:password" in proxy:
            print(f"logins :{json.loads(config['nord_vpn_login'])}")
            nord_vpn_user_pass = random.choice(json.loads(config["nord_vpn_login"]))
            proxy = proxy.replace("username", nord_vpn_user_pass[0]).replace(
                "password", nord_vpn_user_pass[1]
            )
            proxies = {"https": proxy}
            secured = True
        else:
            proxies = {"http": proxy, "https": proxy}
        headers = {
            "User-Agent": random.choice(USER_AGENTS),
            "X-HackerOne-Research": "elniak",
        }
        response = requests.Session().get(
            url,
            proxies=proxies,
            headers=headers,
            verify=secured,
            allow_redirects=True,
            timeout=config["request_delay"],
        )
        delay = random.uniform(config["current_delay"] - 5, config["current_delay"] + 5)
        time.sleep(delay)  # Wait before retrying
        return response.text
    except requests.exceptions.ProxyError as e:
        cprint(
            f"ProxyError searching for {url} with proxy {proxies}: {e}",
            "red",
            file=sys.stderr,
        )
        delay = random.uniform(config["current_delay"] - 2, config["current_delay"] + 2)
        time.sleep(delay)  # Wait before retrying
        # TODO add backoff timer for delay ?
        return None
    except requests.exceptions.RequestException as e:
        cprint(
            f"RequestException searching for {url} with proxy {proxies}: {e}",
            "red",
            file=sys.stderr,
        )
        delay = random.uniform(config["current_delay"] - 5, config["current_delay"] + 5)
        time.sleep(delay)  # Wait before retrying
        return None


def extract_links(html_content, base_url):
    """
    Extract all href links from the HTML content and convert to absolute URLs.

    Args:
        html_content (str): HTML content of the page.
        base_url (str): The base URL to resolve relative URLs.

    Returns:
        list: List of absolute URLs found in the href attributes and other redirections.
    """
    soup = BeautifulSoup(html_content, "html.parser")
    links = []

    # Extracting href attributes from anchor tags
    for anchor in soup.find_all("a", href=True):
        link = urljoin(base_url, anchor["href"])
        links.append(link)

    # Extracting other potential redirections (meta refresh, script-based)
    for meta in soup.find_all("meta", attrs={"http-equiv": "refresh"}):
        content = meta.get("content", "")
        if "url=" in content:
            link = urljoin(base_url, content.split("url=")[-1])
            links.append(link)

    # Handle common JavaScript redirection patterns
    for script in soup.find_all("script"):
        if script.string:
            js_urls = re.findall(
                r"window\.location\.href\s*=\s*['\"](.*?)['\"]", script.string
            )
            for js_url in js_urls:
                link = urljoin(base_url, js_url)
                links.append(link)

    print(f"Extracted {len(links)} links from {base_url}", file=sys.stderr)
    return links


standard_monitors_size = [
    (1920, 1080),
    (1280, 720),
    (854, 480),
    (640, 360),
    (426, 240),
]


random_drivers = ["webdriver.Chrome", "webdriver.Firefox", "webdriver.Edge"]


def render_js_and_get_text(url, proxy=None):
    """
    Use Selenium to render JavaScript and extract links.

    Args:
        url (str): URL of the page to scrape.

    Returns:
        list: List of absolute URLs found in the rendered HTML.
    """
    cprint(f"Rendering JavaScript for {url}", color="yellow", file=sys.stderr)
    try:
        options = Options()
        options.headless = True

        # options.binary_location = "/snap/bin/firefox"
        size = random.choice(standard_monitors_size)
        cprint(f"Selenium Using Size - {size}", color="yellow", file=sys.stderr)
        options.add_argument(f"window-size={size[0]},{size[1]}")
        if proxy:
            options.add_argument(f"--proxy-server={proxy}")
        # service=Service('/path/to/chromedriver'),
        driver = webdriver.Chrome(options=options)  # TODO use edge driver
        # driver = webdriver.Firefox(options=options)

        headers = {
            "User-Agent": random.choice(USER_AGENTS),
        }

        # Create a request interceptor
        def interceptor(request):
            request.headers = headers

        # Set the interceptor on the driver
        driver.request_interceptor = interceptor

        # driver.set_window_size(size[0],size[1])

        driver.get(url)

        by_pass_captcha(driver)

        WebDriverWait(driver, 20).until(
            EC.presence_of_element_located((By.TAG_NAME, "body"))
        )
        html_content = driver.page_source
        driver.quit()
    except Exception as e:
        cprint(f"Error rendering JS for {url}: {e}", color="red", file=sys.stderr)
        # return []
        html_content = ""
    finally:
        return html_content


def render_js_and_extract_links(url, html_content_get, proxy=None):
    """
    Use Selenium to render JavaScript and extract links.

    Args:
        url (str): URL of the page to scrape.

    Returns:
        list: List of absolute URLs found in the rendered HTML.
    """
    cprint(f"Rendering JavaScript for {url}", color="yellow", file=sys.stderr)
    try:
        options = Options()
        options.headless = True
        if proxy:
            options.add_argument(f"--proxy-server={proxy}")
        # service=Service('/path/to/chromedriver'),
        driver = webdriver.Chrome(options=options)
        driver.get(url)
        WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.TAG_NAME, "body"))
        )
        html_content = driver.page_source
        driver.quit()
    except Exception as e:
        cprint(f"Error rendering JS for {url}: {e}", color="red", file=sys.stderr)
        # return []
        html_content = ""
    return extract_links(html_content, url) + extract_links(html_content_get, url)


def scrape_links_from_url(url, proxy=None, headers=None):
    """
    Scrape all potential links and redirections from a given URL.

    Args:
        url (str): The URL to scrape.

    Returns:
        list: List of URLs found in the page.
    """

    cprint(f"Scraping links from {url}", color="yellow", file=sys.stderr)
    html_content = fetch_html_content(url, proxy=proxy, headers=headers)
    if html_content:
        return render_js_and_extract_links(url, html_content, proxy=proxy)
    return []



import json
import random
import re
import sys
import time
from urllib.parse import parse_qs, urlparse

import requests
from termcolor import cprint

from utils.app_config import USER_AGENTS


def headers_403_bypass():
    """_summary_
    https://x.com/bountywriteups/status/1802974307036868838
    Returns:
        _type_: _description_
    """
    headers = {
        "Base-Url": "127.0.0.1",
        "Client-IP": "127.0.0.1",
        "Http-Url": "127.0.0.1",
        "Proxy-Host": "127.0.0.1",
        "Proxy-Url": "127.0.0.1",
        "Real-Ip": "127.0.0.1",
        "Redirect": "127.0.0.1",
        "Referer": "127.0.0.1",
        "Referrer": "127.0.0.1",
        "Refferer": "127.0.0.1",
        "Request-Uri": "127.0.0.1",
        "Uri": "127.0.0.1",
        "Url": "127.0.0.1",
        "X-Client-IP": "127.0.0.1",
        "X-Custom-IP-Authorization": "127.0.0.1",
        "X-Forward-For": "127.0.0.1",
        "X-Forwarded-By": "127.0.0.1",
        "X-Forwarded-For-Original": "127.0.0.1",
        "X-Forwarded-For": "127.0.0.1",
        "X-Forwarded-Host": "127.0.0.1",
        "X-Forwarded-Port": "443",
        "X-Forwarded-Port": "4443",
        "X-Forwarded-Port": "80",
        "X-Forwarded-Port": "8080",
        "X-Forwarded-Port": "8443",
        "X-Forwarded-Scheme": ["http", "https"],
        "X-Forwarded-Server": "127.0.0.1",
        "X-Forwarded": "127.0.0.1",
        "X-Forwarder-For": "127.0.0.1",
        "X-Host": "127.0.0.1",
        "X-Http-Destinationurl": "127.0.0.1",
        "X-Http-Host-Override": "127.0.0.1",
        "X-Original-Remote-Addr": "127.0.0.1",
        "X-Original-Url": "127.0.0.1",
        "X-Originating-IP": "127.0.0.1",
        "X-Proxy-Url": "127.0.0.1",
        "X-Real-Ip": "127.0.0.1",
        "X-Remote-Addr": "127.0.0.1",
        "X-Remote-IP": "127.0.0.1",
        "X-Rewrite-Url": "127.0.0.1",
        "X-True-IP": "127.0.0.1",
    }
    return headers


# Function to check if a given URL has a query string
def has_query_string(url):
    return bool(urlparse.urlparse(url).query)


# Function to inject a payload into a given URL
def inject_payload(url, payload):
    if has_query_string(url):
        url_parts = list(urlparse.urlparse(url))
        query = dict(parse_qs(url_parts[4]))
        for key in query:
            query[key] = f"{query[key]}{payload}"
        url_parts[4] = urlparse.urlencode(query)
        url = urlparse.urlunparse(url_parts)
    else:
        url += f"{payload}"
    return url


def inject_params(url, payload):
    """
    Injects the payload in the parameters and returns a set
    """
    injected_url = set()
    temp_payload = (
        payload.replace("\\n", "$").replace("\\t", "@").replace("\\r", "!")
    )  # saves the payload from the removal of \\n, \\t and \\r
    injected = re.sub(r"=[^?\|&]*", "=" + str(temp_payload), str(url))
    final_payload = injected.replace("$", "\\n").replace("@", "\\t").replace("!", "\\r")
    injected_url.add(final_payload)

    return injected_url


def param_converter(data, url=False):
    if "str" in str(type(data)):
        if url:
            dictized = {}
            parts = data.split("/")[3:]
            for part in parts:
                dictized[part] = part
            return dictized
        else:
            return json.loads(data)
    else:
        if url:
            # url = urlparse(url).scheme + "://" + urlparse(url).netloc
            url = url + "?"
            for key in data.keys():
                url += key + "=" + str(data[key]) + "&"
            return url
        else:
            return json.dumps(data)


def get_url(url, GET):
    if GET:
        return url.split("?")[0]
    else:
        return url


def escaped(position, string):
    usable = string[:position][::-1]
    match = re.search(r"^\\*", usable)
    if match:
        match = match.group()
        if len(match) == 1:
            return True
        elif len(match) % 2 == 0:
            return False
        else:
            return True
    else:
        return False


def is_bad_context(position, non_executable_contexts):
    badContext = ""
    for each in non_executable_contexts:
        if each[0] < position < each[1]:
            badContext = each[2]
            break
    return badContext


xsschecker = "v3dm0s"


def replace_value(mapping, old, new, strategy=None):
    """
    Replace old values with new ones following dict strategy.

    The parameter strategy is None per default for inplace operation.
    A copy operation is injected via strateg values like copy.copy
    or copy.deepcopy

    Note: A dict is returned regardless of modifications.
    """
    anotherMap = strategy(mapping) if strategy else mapping
    if old in anotherMap.values():
        for k in anotherMap.keys():
            if anotherMap[k] == old:
                anotherMap[k] = new
    return anotherMap


def js_extractor(response):
    """Extract js files from the response body"""
    scripts = []
    matches = re.findall(r"<(?:script|SCRIPT).*?(?:src|SRC)=([^\s>]+)", response)
    for match in matches:
        if xsschecker in match:
            match = match.replace("'", "").replace('"', "").replace("`", "")
            scripts.append(match)
    return scripts


def stripper(string, substring, direction="right"):
    done = False
    strippedString = ""
    if direction == "right":
        string = string[::-1]
    for char in string:
        if char == substring and not done:
            done = True
        else:
            strippedString += char
    if direction == "right":
        strippedString = strippedString[::-1]
    return strippedString


def fill_holes(original, new):
    filler = 0
    filled = []
    for x, y in zip(original, new):
        if int(x) == (y + filler):
            filled.append(y)
        else:
            filled.extend([0, y])
            filler += int(x) - y
    return filled


def de_json(data):
    return data.replace("\\\\", "\\")


def handle_anchor(parent_url, url):
    """
    Constructs a complete URL based on the parent URL and the given URL.

    Args:
        parent_url (str): The parent URL.
        url (str): The URL to be handled.

    Returns:
        str: The complete URL.

    """
    scheme = urlparse(parent_url).scheme
    if url[:4] == "http":
        return url
    elif url[:2] == "//":
        return scheme + ":" + url
    elif url.startswith("/"):
        host = urlparse(parent_url).netloc
        scheme = urlparse(parent_url).scheme
        parent_url = scheme + "://" + host
        return parent_url + url
    elif parent_url.endswith("/"):
        return parent_url + url
    else:
        return parent_url + "/" + url


def get_params(url, data, GET):
    """Get the parameters from the given URL and data.

    This function parses the URL and data to extract the parameters and returns them as a dictionary.

    Args:
        url (str): The URL string.
        data (str): The data string.
        GET (str): The GET string.

    Returns:
        dict: A dictionary containing the extracted parameters.
    """
    params = {}
    if "?" in url and "=" in url:
        data = url.split("?")[1]
        if data[:1] == "?":
            data = data[1:]
    elif data:
        if False:  # TODO: get_var('jsonData') or get_var('path'):
            params = data
        else:
            try:
                params = json.loads(data.replace("'", '"'))
                return params
            except json.decoder.JSONDecodeError:
                pass
    else:
        return None
    if not params:
        parts = data.split("&")
        for part in parts:
            each = part.split("=")
            if len(each) < 2:
                each.append("")
            try:
                params[each[0]] = each[1]
            except IndexError:
                params = None
    return params


def start_request(
    proxies,
    config,
    is_json=False,
    GET=False,
    data=None,
    headers=None,
    params=None,
    base_url=None,
    secured=False,
    cookies=None,
    session=None,
    bypassed_403=False,
):
    """
    Send a HTTP request to the specified URL.

    Args:
        proxies (dict): A dictionary of proxy settings.
        config (dict): A dictionary of configuration settings.
        is_json (bool, optional): Indicates whether the request data is in JSON format. Defaults to False.
        GET (bool, optional): Indicates whether the request method is GET. Defaults to False.
        data (dict or str, optional): The request data. Defaults to None.
        headers (dict, optional): A dictionary of request headers. Defaults to None.
        params (dict, optional): A dictionary of request parameters. Defaults to None.
        base_url (str, optional): The base URL for the request. Defaults to None.
        secured (bool, optional): Indicates whether the request should be verified using SSL/TLS. Defaults to False.
        cookies (dict, optional): A dictionary of request cookies. Defaults to None.
        session (object, optional): A session object to use for the request. Defaults to None.

    Returns:
        object: The response object returned by the request.
    """
    if session:
        requester = session
    else:
        requester = requests
    try:
        if GET:
            response = requester.get(
                base_url,
                headers=headers,
                params=params,
                allow_redirects=True,
                proxies=proxies,
                # cookies = {'CONSENT' : 'YES+'},
                cookies=cookies,  # FOR EU USERS -> ANNOYING to parse
                verify=secured,  # TODO add parameter for that
                timeout=config["request_delay"],
            )
        elif is_json:
            response = requester.post(
                base_url,
                json=data[0],
                headers=headers,
                timeout=config["request_delay"],
                verify=secured,
                cookies=cookies,  # FOR EU USERS
                proxies=proxies,
            )
        else:
            response = requester.post(
                base_url,
                data=data[0],
                headers=headers,
                timeout=config["request_delay"],
                verify=secured,
                cookies=cookies,  # FOR EU USERS
                proxies=proxies,
            )

        # Parse Google response
        if response.status_code != 200:
            cprint(
                f"Error in request ... - status code = {response.status_code}",
                color="red",
                file=sys.stderr,
            )
            if response.status_code == 429:
                # delay = random.uniform(LONG_DELAY-5, LONG_DELAY+5)
                # time.sleep(delay)  # Wait before retrying
                retry_after = int(response.headers.get("Retry-After", 60))
                cprint(
                    f"Retry after {retry_after} secs ...",
                    "red",
                    file=sys.stderr,
                )
                time.sleep(retry_after)
            elif response.status_code == 403:
                # TODO with headers_403_bypass()
                if not bypassed_403:
                    cprint(
                        "403 Forbidden - Trying to bypass ...",
                        "yellow",
                        file=sys.stderr,
                    )
                    delay = random.uniform(
                        config["current_delay"] - 5, config["current_delay"] + 5
                    )
                    time.sleep(delay)  # Wait before retrying
                    response = start_request(
                        proxies=proxies,
                        base_url=base_url,
                        params=params,
                        headers=headers_403_bypass(),
                        secured=secured,
                        GET=GET,
                        config=config,
                        bypassed_403=True,
                    )
                else:
                    cprint(
                        "WAF is dropping suspicious requests. Scanning will continue after 10 minutes.",
                        color="red",
                        file=sys.stderr,
                    )
                    time.sleep(config["waf_delay"])
            else:
                cprint(
                    f"Error in request ... - status code = {response.status_code}",
                    color="red",
                    file=sys.stderr,
                )
                delay = random.uniform(
                    config["current_delay"] - 5, config["current_delay"] + 5
                )
                time.sleep(delay)  # Wait before retrying
            return response
        elif (
            "did not match any documents" in response.text
            and "Your search -" in response.text
        ):
            cprint(
                f"No results found for {params['q'] if  params and 'q' in params else params} with proxy {proxies}",
                "yellow",
                file=sys.stderr,
            )
            delay = random.uniform(
                config["current_delay"] - 5, config["current_delay"] + 5
            )
            time.sleep(delay)  # Wait before retrying
            return response
        # Placeholder for URL extraction logic
        cprint("Request successful - 200", "green", file=sys.stderr)
        delay = random.uniform(config["current_delay"] - 5, config["current_delay"] + 5)
        time.sleep(delay)  # Wait before retrying
        return response  # Return the category and a placeholder result
    except requests.exceptions.ProxyError as e:
        cprint(
            f"ProxyError searching for {params['q'] if  params and 'q' in params else params} with proxy {proxies}: {e}",
            "red",
            file=sys.stderr,
        )
        delay = random.uniform(config["current_delay"] - 2, config["current_delay"] + 2)
        time.sleep(delay)  # Wait before retrying
        return e.response
    except requests.exceptions.RequestException as e:
        cprint(
            f"RequestException searching for {params['q'] if  params and 'q' in params else params} with proxy {proxies}: {e}",
            "red",
            file=sys.stderr,
        )
        delay = random.uniform(config["current_delay"] - 2, config["current_delay"] + 2)
        time.sleep(delay)  # Wait before retrying
        return e.response



#########################################################################################
# Github Dorking and searching functions
#########################################################################################

# Token Round Robin for Sequential Requests
from datetime import datetime
import random
import time

from bs4 import BeautifulSoup
import requests
from termcolor import cprint

from utils.app_config import *

token_index = 0

# GitHub Dorking
GITHUB_API_URL = "https://api.github.com"
TOKENS_LIST = ["your_github_token"]  # Add your GitHub tokens here
DORK_LIST = ["example_dork1", "example_dork2"]  # Add your dorks here
QUERIES_LIST = ["example_query"]  # Add your queries here
ORGANIZATIONS_LIST = ["example_organization"]  # Add your organizations here
USERS_LIST = ["example_user"]  # Add your users here


def token_round_robin():
    global token_index
    token = TOKENS_LIST[token_index]
    token_index = (token_index + 1) % len(TOKENS_LIST)
    return token


# URL Encoding Function
def __urlencode(str):
    return str.replace(":", "%3A").replace('"', "%22").replace(" ", "+")


def rate_limit_handler(headers):
    if "X-RateLimit-Reset" in headers:
        reset_time = datetime.fromtimestamp(int(headers["X-RateLimit-Reset"]))
        wait_time = (
            reset_time - datetime.now()
        ).total_seconds() + 10  # Adding 10 seconds buffer
        print(f"Rate limit hit. Waiting for {wait_time} seconds.")
        time.sleep(wait_time)


def get_rate_limit_status():
    headers = {"Authorization": f"token {token_round_robin()}"}
    response = requests.get(f"{GITHUB_API_URL}/rate_limit", headers=headers)
    if response.status_code == 200:
        rate_limit = response.json()["rate"]
        print(
            f"Limit: {rate_limit['limit']}, Remaining: {rate_limit['remaining']}, Reset: {datetime.fromtimestamp(rate_limit['reset'])}"
        )
    else:
        print("Failed to fetch rate limit status")


# TODO
def github_search_with_proxy(dork_tuple, proxy, category, retries=3, advanced=False):
    # TODO advanced search
    try:
        query, extension = dork_tuple
    except ValueError:
        query = dork_tuple
        extension = ""
    full_query = f"{query} {extension}".strip()
    base_url = f"{GITHUB_API_URL}/search/code?q=" + __urlencode(extension + " " + query)
    headers = {
        "User-Agent": random.choice(USER_AGENTS),
        "Authorization": f"token {token_round_robin()}",
    }
    proxies = {"http": proxy, "https": proxy}
    urls = []
    for _ in range(retries):
        try:
            cprint(
                f"Searching for {full_query} ({category}) with proxy {proxy}...",
                "yellow",
                file=sys.stderr,
            )
            response = requests.get(
                base_url, headers=headers, params=params, proxies=proxies, timeout=10
            )

            # Parse
            soup = BeautifulSoup(response.text, "html.parser")
            result_block = soup.find_all("div", attrs={"class": "g"})
            for result in result_block:
                # Find link, title, description
                link = result.find("a", href=True)
                title = result.find("h3")
                description_box = result.find("div", {"style": "-webkit-line-clamp:2"})
                if description_box:
                    description = description_box.text
                    if link and title and description:
                        if advanced:
                            urls.append(
                                SearchResult(link["href"], title.text, description)
                            )
                        else:
                            urls.append(link["href"])

            # Placeholder for URL extraction logic
            return category, urls  # Return the category and a placeholder result
        except requests.exceptions.RequestException as e:
            # cprint(f"Error searching for {full_query} with proxy {proxy}: {e}", 'red', file=sys.stderr)
            time.sleep(2)  # Wait before retrying

    return category, None  # Indicate failure after retries


# def launch_github_dorks_and_search_attack(extension=DEFAULT_EXTENSION, total_output=DEFAULT_TOTAL_OUTPUT, page_no=DEFAULT_PAGE_NO, proxies=None):
#     pass



import sys
from termcolor import cprint
from vpn_proxies.nordvpn_switcher.nordvpn_switch import (
    initialize_VPN,
    rotate_VPN,
    terminate_VPN,
)


def setup_vpn(config):
    """Set up VPN based on the provided configuration.

    Args:
        config (dict): A dictionary containing the VPN configuration.

    Raises:
        NotImplementedError: Raised if NordVPN is not supported in this version.
    """
    if config["use_nordvpn"]:
        raise NotImplementedError(
            "NordVPN is not supported in this version - Error in library"
        )
        if len(config["nord_vpn_login"]) > 0:
            try:
                initialize_VPN(save=1, area_input=["complete rotation"])
            except Exception as e:
                cprint(
                    f"VPN initialization error: {e}",
                    "red",
                    file=sys.stderr,
                )
                config["use_nordvpn"] = False
        else:
            cprint(
                "You need to provide NordVPN credentials to use VPN",
                "red",
                file=sys.stderr,
            )
            config["use_nordvpn"] = False


def change_vpn(time=300):
    rotate_VPN()
    time.sleep(time)



import csv
import json
import os
import sys

from termcolor import cprint

import threading

google_dorking_results = []


LOCKS = {
    "dorking": threading.Lock(),
}
#########################################################################################
# File writing functions
#########################################################################################

# Define file paths
# Initialize locks for thread-safe file writing
# TODO make more modular


def get_processed_dorks(settings):
    """
    Reads the experiment CSV file to get the list of processed dorks.
    """
    processed_dorks = set()

    if os.path.exists(settings["dorking_csv"]):
        with open(settings["dorking_csv"], mode="r", newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                processed_dorks.add(row["dork"])

    return processed_dorks


def get_last_processed_ids(settings):
    """
    Get the last processed dork_id, link_id, and attack_id from the CSV file.
    """
    last_dork_id = 0
    last_link_id = 0
    last_attack_id = 0

    if os.path.exists(settings["dorking_csv"]):
        with open(settings["dorking_csv"], mode="r", newline="") as file:
            reader = csv.DictReader(file)
            for row in reader:
                last_dork_id = int(row["dork_id"])
                last_link_id = int(row["link_id"])
                last_attack_id = int(row["attack_id"])

    return last_dork_id, last_link_id, last_attack_id

# Thread-safe addition to results lists
def save_dorking_query(result, settings):
    """
    Safely adds results to the single experiment CSV file with tracking IDs.
    """
    dork_id, category, urls, dork = result
    with LOCKS["dorking"]:
        with open(settings["dorking_csv"], mode="a", newline="") as file:
            writer = csv.writer(file)
            _, link_id, last_attack_id = get_last_processed_ids(settings)
            link_id += 1  # Increment link_id for next link
            if urls:
                cprint(
                    f"Adding {len(urls)} URLs to experiment list...",
                    "blue",
                    file=sys.stderr,
                )
                for url in urls:
                    if url and "https://www.google.com/sorry/" not in url:
                        attack_id = (
                            last_attack_id  # Start attack_id from 1 for each link
                        )
                        row = [
                            dork_id,
                            link_id,
                            attack_id,
                            category,
                            url,
                            dork,
                            "yes",
                            "",
                        ]  # Success and payload columns are initially empty
                        # if settings["do_dorking_github"] and category == "github":
                        #     row.append("no")
                        writer.writerow(row)
                        cprint(
                            f"Added {url} to experiment list under category {category}",
                            "blue",
                            file=sys.stderr,
                        )
                        attack_id += 1  # Increment attack_id for next attack
                    else:
                        cprint(
                            f"Google blocked us from accessing {url}",
                            "red",
                            file=sys.stderr,
                        )
                link_id += 1  # Increment link_id for next link
            else:
                # Write a row indicating no URLs found for this dork
                row = [
                    dork_id,
                    link_id,
                    last_attack_id,
                    category,
                    "",
                    dork,
                    "no",
                    "",
                ]  # No URLs found
                # if settings["do_dorking_github"]:
                #     row.append("no")
                
                writer.writerow(row)
                cprint(f"No URLs found for {category} dorks...", "red", file=sys.stderr)



# /usr/bin/python3
import argparse
import sys

# TO get stacktrace in case of segfault
# gdb pypy3; run bounty_drive.py
# import faulthandler
# faulthandler.enable(file=sys.stderr, all_threads=True)

import utils.logger
from utils.logger import *


from termcolor import cprint
import os
import csv
import concurrent.futures

from tqdm import tqdm

from dorks.search_engine_dorking import (
    google_search_with_proxy,
    launch_google_dorks_and_search_attack,
)

from dorks.github_dorking import *

from reporting.results_manager import (
    get_last_processed_ids,
    get_processed_dorks,
)
from vpn_proxies.proxies_manager import round_robin_proxies, setup_proxies
from vpn_proxies.vpn_manager import setup_vpn


from utils.banner import *
from utils.banner import load_animation


import configparser

os.system("clear")

#########################################################################################
# Main function
#########################################################################################


def create_argument_parser():
    """
    Creates an argument parser for command-line arguments.
    
    Returns:
    - parser: the argument parser
    """
    parser = argparse.ArgumentParser(description="Configuration and Argument Parser")

    parser.add_argument("--config", type=str, required=True, help="Path to the configuration file")

    # [Settings] section
    parser.add_argument("--extension", type=str, help="Extension")
    parser.add_argument("--subdomain", type=bool, help="Use subdomain")
    parser.add_argument("--do_web_scap", type=bool, help="Do web scraping")
    parser.add_argument("--target_file", type=str, help="Target file")
    parser.add_argument("--exclusion_file", type=str, help="Exclusion file")
    parser.add_argument("--target_login", type=str, nargs='*', help="Target login")
    parser.add_argument("--logging", type=str, help="Logging level")
    parser.add_argument("--max_thread", type=int, help="Maximum number of threads")
    parser.add_argument("--runtime_save", type=bool, help="Runtime save")
    parser.add_argument("--keyboard_interrupt_save", type=bool, help="Keyboard interrupt save")

    # [Bounty] section
    parser.add_argument("--need_specific_user_agent", type=bool, help="Need specific user agent")
    parser.add_argument("--target_user_agent", type=str, help="Target user agent")
    parser.add_argument("--hackerone_username", type=str, help="HackerOne username")

    # [GoogleDorking] section
    parser.add_argument("--do_dorking_google", type=bool, help="Do Google dorking")
    parser.add_argument("--total_output", type=int, help="Total output")
    parser.add_argument("--page_no", type=int, help="Page number")
    parser.add_argument("--default_total_output", type=int, help="Default total output")
    parser.add_argument("--default_page_no", type=int, help="Default page number")
    parser.add_argument("--lang", type=str, help="Language")
    parser.add_argument("--use_selenium", type=bool, help="Use Selenium")
    parser.add_argument("--do_xss", type=bool, help="Do XSS")
    parser.add_argument("--do_sqli", type=bool, help="Do SQLi")
    
    # [GithubDorking] section
    parser.add_argument("--do_dorking_github", type=bool, help="Do GitHub dorking")

    # [ShodanDorking] section
    parser.add_argument("--do_dorking_shodan", type=bool, help="Do Shodan dorking")

    # [Proxy] section
    parser.add_argument("--use_proxy", type=bool, help="Use proxy")
    parser.add_argument("--use_free_proxy_file", type=bool, help="Use free proxy file")
    parser.add_argument("--use_free_proxy", type=bool, help="Use free proxy")
    parser.add_argument("--use_nordvpn_proxy", type=bool, help="Use NordVPN proxy")
    parser.add_argument("--proxies", type=str, nargs='*', help="Proxies")
    parser.add_argument("--proxy_mean_delay", type=int, help="Proxy mean delay")
    parser.add_argument("--proxy_factor", type=int, help="Proxy factor")

    # [VPN] section
    parser.add_argument("--use_vpn", type=bool, help="Use VPN")
    parser.add_argument("--use_nordvpn", type=bool, help="Use NordVPN")
    parser.add_argument("--nord_vpn_login", type=str, nargs='*', help="NordVPN login")

    # [Tor] section
    parser.add_argument("--use_tor", type=bool, help="Use Tor")

    # [Delay] section
    parser.add_argument("--initial_delay", type=int, help="Initial delay")
    parser.add_argument("--delay_factor", type=int, help="Delay factor")
    parser.add_argument("--long_delay", type=int, help="Long delay")
    parser.add_argument("--max_delay", type=int, help="Max delay")
    parser.add_argument("--request_delay", type=int, help="Request delay")
    parser.add_argument("--waf_delay", type=int, help="WAF delay")

    # [Rate] section
    parser.add_argument("--rate_per_minute", type=int, help="Rate per minute")
    parser.add_argument("--current_delay", type=int, help="Current delay")

    return parser
    
def read_config(file_path):
    """
    Reads the configuration file and returns the settings as a dictionary.
    """
    config = configparser.ConfigParser()
    config.read(file_path)

    settings = {
        # Settings
        "extension": config["Settings"].get("extension"),
        "subdomain": config["Settings"].getboolean("subdomain"),
        "do_web_scrap": config["Settings"].getboolean("do_web_scrap"),
        "target_file": config["Settings"].get("target_file"),
        "max_thread": config["Settings"].getint("max_thread", 30),
        "logging": config["Settings"].get("logging", "DEBUG"),
        "runtime_save": config["Settings"].getboolean("runtime_save"),
        "keyboard_interrupt_save": config["Settings"].getboolean(
            "keyboard_interrupt_save"
        ),
        # Bounty
        "need_specific_user_agent": config["Bounty"].getboolean(
            "need_specific_user_agent"
        ),
        "target_user_agent": config["Bounty"].get("target_user_agent"),
        "hackerone_username": config["Bounty"].get("hackerone_username"),
        # Google Dorking
        "do_dorking_google": config["GoogleDorking"].getboolean("do_dorking_google"),
        "total_output": config["GoogleDorking"].getint("total_output"),
        "page_no": config["GoogleDorking"].getint("page_no"),
        "default_total_output": config["GoogleDorking"].getint("default_total_output"),
        "default_page_no": config["GoogleDorking"].getint("default_page_no"),
        "lang": config["GoogleDorking"].get("lang"),
        "use_selenium": config["GoogleDorking"].getboolean("use_selenium"),
        "do_xss": config["GoogleDorking"].getboolean("do_xss"),
        "do_sqli": config["GoogleDorking"].getboolean("do_sqli"),
        # Github Dorking
        "do_dorking_github": config["GithubDorking"].getboolean("do_dorking_github"),
        # Github Dorking
        "do_dorking_shodan": config["GithubDorking"].getboolean("do_dorking_github"),
        # Proxy
        "use_proxy": config["Proxy"].getboolean("use_proxy"),
        "use_free_proxy_file": config["Proxy"].getboolean("use_free_proxy_file"),
        "use_free_proxy": config["Proxy"].getboolean("use_free_proxy"),
        "use_nordvpn_proxy": config["Proxy"].getboolean("use_nordvpn_proxy"),
        "proxies": config["Proxy"].get("proxies"),
        "proxy_mean_delay": config["Proxy"].getint("proxy_mean_delay"),
        "proxy_factor": config["Proxy"].getint("proxy_factor"),
        # VPN
        "use_vpn": config["VPN"].getboolean("use_vpn"),
        "use_nordvpn": config["VPN"].getboolean("use_nordvpn"),
        "nord_vpn_login": config["VPN"].get("nord_vpn_login"),
        # Delay
        "initial_delay": config["Delay"].getint("initial_delay"),
        "delay_factor": config["Delay"].getint("delay_factor"),
        "long_delay": config["Delay"].getint("long_delay"),
        "max_delay": config["Delay"].getint("max_delay"),
        "request_delay": config["Delay"].getint("request_delay"),
        "waf_delay": config["Delay"].getint("waf_delay"),
        # Rate
        "rate_per_minute": config["Rate"].getint("rate_per_minute"),
        "current_delay": config["Rate"].getint("current_delay"),
    }

    return settings


def get_user_input(config_file="configs/config.ini"):
    """
    Collects user input from a configuration file.

    Args:
        config_file (str): Path to the configuration file. Default is "configs/config.ini".

    Returns:
        tuple: A tuple containing the following values:
            - config (dict): The configuration settings read from the file.
            - last_dork_id (int): The last processed dork ID.
            - last_link_id (int): The last processed link ID.
            - last_attack_id (int): The last processed attack ID.
            - categories (list): A list of categories.

    Raises:
        FileNotFoundError: If the specified configuration file does not exist.
    """
    config = {}
    if args.config:
        config = read_config(args.config)
    cli_args = vars(args)

    # Override config settings with command-line arguments if provided
    for key in cli_args:
        if cli_args[key] is not None:
            config[key] = cli_args[key]

    categories = []

    # Define headers based on enabled parameters
    setup_experiment_folder(config, categories)

    cprint(
        f"-Extension: {config['extension']}\n-Total Output: {config['total_output']}\n-Page No: {config['page_no']}\n-Do Google Dorking: {config['do_dorking_google']}\n-Do Github Dorking {config['do_dorking_github']}\n-Domain: {config['subdomain']}\n-Use Proxy: {config['use_proxy']}",
        "blue",
        file=sys.stderr,
    )

    if config["use_proxy"]:
        setup_proxies(config)

    if config["use_vpn"]:
        setup_vpn(config)

    last_dork_id, last_link_id, last_attack_id = get_last_processed_ids(config)

    if config["subdomain"]:
        proxies = config["proxies"]
        if config["use_proxy"] and len(proxies) == 0:
            cprint(
                f"Using proxies -> you should have at least one UP",
                "red",
                file=sys.stderr,
            )
            exit()

        if not config["use_proxy"]:
            proxies = [None]

        proxy_cycle = round_robin_proxies(proxies)

        saved_total_output = config["total_output"]
        current_total_output = 10
        config["total_output"] = current_total_output
        search_tasks_with_proxy = []

        number_of_worker = 30 
        cprint(f"Number of workers: {number_of_worker}", "blue", file=sys.stderr)

        with open(config["target_file"], "r") as file:
            subdomain_list = file.read().splitlines()
            if len(subdomain_list) >= last_dork_id:
                for domain in subdomain_list:
                    processed = False
                    for category in categories:
                        with open(config["dorking_csv"], mode="r", newline="") as file:
                            reader = csv.DictReader(file)
                            for row in reader:
                                if domain in row["dork"]:
                                    processed = True

                        if not processed:
                            proxy = next(proxy_cycle)
                            search_tasks_with_proxy.append(
                                {
                                    "dork": "",
                                    "proxy": proxy,
                                    "category": category,
                                    "domain": domain,
                                }
                            )
                            cprint(
                                f"Initial Dorking search for based targets {domain} - {category}",
                                "yellow",
                                file=sys.stderr,
                            )
                        else:
                            cprint(
                                f"Already initialized Dorking search for based targets {domain} - {category}",
                                color="cyan",
                                file=sys.stderr,
                            )

            processed_dorks = get_processed_dorks(config)
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=number_of_worker
            ) as executor:
                future_to_search = {
                    executor.submit(
                        google_search_with_proxy,
                        task["dork"],
                        task["proxy"],
                        task["category"],
                        config,
                        task["domain"],
                        processed_dorks,
                    ): task
                    for task in search_tasks_with_proxy
                }
                for future in tqdm(
                    concurrent.futures.as_completed(future_to_search),
                    total=len(future_to_search),
                    desc="Initializing Dorking of targets",
                    unit="site",
                ):
                    task = future_to_search[future]
                    # try:
                    last_dork_id = future.result()

            config["subdomain"] = subdomain_list
    else:
        config["subdomain"] = [None]

    config["total_output"] = saved_total_output

    return config, last_dork_id, last_link_id, last_attack_id, categories


def setup_experiment_folder(config, categories):
    folder_name = os.path.join(
        "outputs/reports/", config["target_file"].split("_")[-1].replace(".txt", "")
    )
    config["experiment_folder"] = folder_name
    if not os.path.exists(folder_name):
        cprint(f"Creating folder {folder_name}", "blue", file=sys.stderr)
        os.makedirs(folder_name)
    setup_csv(config, categories, folder_name)


def setup_csv(config, categories, folder_name):
    """Set up the CSV file for storing experiment results.

    This function creates the necessary CSV file and writes the headers based on the provided configuration.

    Args:
        config (dict): A dictionary containing the configuration settings.
        categories (list): A list of categories to be included in the CSV headers.

    """
    csv_headers = [
        "dork_id",
        "link_id",
        "attack_id",
        "category",
        "url",
        "dork",
        "success",
        "payload",
    ]
    config["dorking_csv"] = os.path.join(
        folder_name, folder_name.split("/")[-1] + "_dorking.csv"
    )
    if config["do_dorking_github"]:
        csv_headers.append("github_success")
    if config["do_xss"]:
        categories.append("xss")
    if config["do_sqli"]:
        categories.append("sqli")
    if (
        not os.path.exists(config["dorking_csv"])
        or os.path.getsize(config["dorking_csv"]) == 0
    ):
        with open(config["dorking_csv"], mode="a", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(csv_headers)


if __name__ == "__main__":
    try:
        load_animation()

        # TODO add worker/master processes to handle multiple tasks and be faster

        # Create argument parser
        parser = create_argument_parser()

        # Parse command-line arguments
        args = parser.parse_args()

        (
            config,
            last_dork_id,
            last_link_id,
            last_attack_id,
            categories,
        ) = get_user_input(args)
        

        if config["do_dorking_google"]:
            cprint(
                "\nStarting Google dorking scan phase...\n", "blue", file=sys.stderr
            )
            launch_google_dorks_and_search_attack(config, categories)

        if config["do_dorking_github"]:
            cprint(
                "\nStarting Github dorking scan phase...\n", "blue", file=sys.stderr
            )
            raise NotImplementedError("Github dorking scan phase not implemented yet")
            launch_github_dorks_and_search_attack(config, categories)

        if config["do_dorking_shodan"]:
            cprint(
                "\nStarting Shodan dorking scan phase...\n", "blue", file=sys.stderr
            )
            raise NotImplementedError("Shodan dorking scan phase not implemented yet")
            launch_shodan_dorks_and_search_attack(config, categories)
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        cprint(f"{exc_type}, {fname}, {exc_tb.tb_lineno}", "red", file=sys.stderr)
        cprint(f"Error: {e}", "red", file=sys.stderr)
    except KeyboardInterrupt:
        cprint("Exiting...", "red", file=sys.stderr)
        # TODO save progress
    finally:
        sys.stderr = orig_stdout
        f.close()



#################
# import packages#
#################
# 1.system and terminal/cmd
import os
from os import path
import platform
import subprocess
from subprocess import check_output, DEVNULL

# 2.utilities (randomisations etc)
import psutil
import random
import re
import time

# 3.scraping
import urllib
from bs4 import BeautifulSoup

# TODO USE this package or reimplement ?
from random_user_agent.user_agent import UserAgent
from random_user_agent.params import SoftwareName, OperatingSystem
import requests

# 4.file formats
import json

# 5.package dependencies
import importlib.resources as pkg_resources
from vpn_proxies.nordvpn_switcher import NordVPN_options

##################################
# retrieve useragents for scraping#
##################################
software_names = [SoftwareName.CHROME.value]
operating_systems = [OperatingSystem.WINDOWS.value, OperatingSystem.LINUX.value]
user_agent_rotator = UserAgent(
    software_names=software_names, operating_systems=operating_systems, limit=100
)


##################
# helper functions#
##################
def additional_settings_linux(additional_settings):
    try:
        additional_setting_execute = str(check_output(additional_settings))
    except:
        additional_setting_execute = "error"
    else:
        pass

    if "successfully" in additional_setting_execute:
        settings_input_message = "\nDone! Anything else?\n"
    elif additional_setting_execute == "error":
        settings_input_message = "\n\x1b[93mSomething went wrong. Please consult some examples by typing 'help'\x1b[0m\n"
    elif "already" in additional_setting_execute:
        settings_input_message = (
            "\nThis setting has already been executed! Anything else?\n"
        )
    else:
        settings_input_message = (
            "\n\x1b[93mNordVPN throws an unexpected message, namely:\n"
            + additional_setting_execute
            + "\nTry something different.\x1b[0m\n"
        )
    return settings_input_message


def saved_settings_check():
    print("\33[33mTrying to load saved settings...\33[0m")
    try:
        instructions = json.load(
            open("vpn_proxies/nordvpn_switcher/settings_nordvpn.txt")
        )
    except FileNotFoundError:
        raise Exception(
            "\n\nSaved settings not found.\n"
            "Run initialize_VPN() first and save the settings on your hard drive or store it into a Python variable."
        )
    else:
        print("\33[33mSaved settings loaded!\n\33[0m")
    return instructions


def set_headers(user_agent_rotator):
    useragent_pick = user_agent_rotator.get_random_user_agent()
    headers = {
        "User-Agent": useragent_pick,
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "Accept-Charset": "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
        "Accept-Encoding": "none",
        "cache-control": "max-age=0",
        "Accept-Language": "en-US,en;q=0.8",
        "Connection": "keep-alive",
    }
    return headers


def get_ip():
    headers = set_headers(user_agent_rotator)
    ip_check_websites = ["http://ip4only.me/api/", "https://api64.ipify.org/"]
    website_pick = random.choice(ip_check_websites)
    request_currentip = urllib.request.Request(url=website_pick, headers=headers)
    ip = urllib.request.urlopen(request_currentip).read().decode("utf-8")
    if website_pick == "http://ip4only.me/api/":
        ip = re.search("IPv4,(.*?),Remaining", ip).group(1)
    return ip


def get_nordvpn_servers():
    serverlist = BeautifulSoup(
        requests.get("https://api.nordvpn.com/v1/servers").content, "html.parser"
    )
    site_json = json.loads(serverlist.text)

    filtered_servers = {key: [] for key in ["windows_names", "linux_names"]}
    for specific_dict in site_json:
        try:
            if specific_dict["status"] == "online":
                filtered_servers["windows_names"].append(specific_dict["name"])
                filtered_servers["linux_names"].append(
                    specific_dict["hostname"].split(".")[0]
                )
        except IndexError:
            pass
    return filtered_servers


###############
# intialize vpn#
###############
def initialize_VPN(stored_settings=0, save=0, area_input=None, skip_settings=0):

    ###load stored settings if needed and set input_needed variables to zero if settings are provided###
    windows_pause = 3
    additional_settings_needed = 1
    additional_settings_list = list()
    if stored_settings == 1:
        instructions = saved_settings_check()
        additional_settings_needed = 0
        input_needed = 0
    elif area_input is not None:
        input_needed = 2
        windows_pause = 8
    else:
        input_needed = 1

    ###performing system check###
    opsys = platform.system()

    ##windows##
    if opsys == "Windows":
        print(
            "\33[33mYou're using Windows.\n"
            "Performing system check...\n"
            "###########################\n\33[0m"
        )
        # seek and set windows installation path#
        option_1_path = "C:/Program Files/NordVPN"
        option_2_path = "C:/Program Files (x86)/NordVPN"
        custom_path = str()
        if path.exists(option_1_path) == True:
            cwd_path = option_1_path
        elif path.exists(option_2_path) == True:
            cwd_path = option_2_path
        else:
            custom_path = input(
                "\x1b[93mIt looks like you've installed NordVPN in an uncommon folder. Would you mind telling me which folder? (e.g. D:/customfolder/nordvpn)\x1b[0m"
            )
            while path.exists(custom_path) == False:
                custom_path = input(
                    "\x1b[93mI'm sorry, but this folder doesn't exist. Please double-check your input.\x1b[0m"
                )
            while os.path.isfile(custom_path + "/NordVPN.exe") == False:
                custom_path = input(
                    "\x1b[93mI'm sorry, but the NordVPN application is not located in this folder. Please double-check your input.\x1b[0m"
                )
            cwd_path = custom_path
        print("NordVPN installation check: \33[92m\N{check mark}\33[0m")

        # check if nordvpn service is already running in the background
        check_service = "nordvpn-service.exe" in (
            p.name() for p in psutil.process_iter()
        )
        if check_service is False:
            raise Exception(
                "NordVPN service hasn't been initialized, please start this service in [task manager] --> [services] and restart your script"
            )
        print("NordVPN service check: \33[92m\N{check mark}\33[0m")

        # start NordVPN app and disconnect from VPN service if necessary#
        print("Opening NordVPN app and disconnecting if necessary...")
        open_nord_win = subprocess.Popen(
            ["nordvpn", "-d"], shell=True, cwd=cwd_path, stdout=DEVNULL
        )
        while ("NordVPN.exe" in (p.name() for p in psutil.process_iter())) == False:
            time.sleep(windows_pause)
        open_nord_win.kill()
        print("NordVPN app launched: \33[92m\N{check mark}\33[0m")
        print("#####################################")

    ##linux##
    elif opsys == "Linux":
        print(
            "\n\33[33mYou're using Linux.\n"
            "Performing system check...\n"
            "###########################\n\33[0m"
        )

        # check if nordvpn is installed on linux#
        check_nord_linux = check_output(["nordvpn"])
        if len(check_nord_linux) > 0:
            print("NordVPN installation check: \33[92m\N{check mark}\33[0m")
        else:
            raise Exception(
                "NordVPN is not installed on your Linux machine.\n"
                "Follow instructions on shorturl.at/ioDQ2 to install the NordVpn app."
            )

        # check if user is logged in. If not, ask for credentials and log in or use credentials from stored settings if available.#
        try:
            check_output(["nordvpn", "account"])
        except:
            login_needed = 1
            while login_needed == 1:
                login_message = input(
                    "\n\033[34mYou are not logged in. Please provide your credentials in the form of LOGIN/PASSWORD\n\033[0m"
                )
                try:
                    if instructions["credentials"] in locals():
                        credentials = stored_settings["credentials"]
                    else:
                        credentials = login_message
                except:
                    credentials = login_message
                finally:
                    try:
                        login = credentials.split("/")[0]
                        password = credentials.split("/")[1]
                    except IndexError:
                        error_login = input(
                            "\n\033[34mYou have provided your credentials in the wrong format. Press enter and please try again.\n"
                            "Your input should look something like this: name@gmail.com/password\033[0m"
                        )
                    else:
                        login_needed = 0
            try:
                login_nordvpn = check_output(
                    ["nordvpn", "login", "-username", login, "-password", password]
                )
            except subprocess.CalledProcessError:
                raise Exception("\nSorry,something went wrong while trying to log in\n")
            if "Welcome" in str(login_nordvpn):
                print("\n\n\033[34mLogin successful!\n\033[0m\n")
                pass
            else:
                raise Exception(
                    "\nSorry, NordVPN throws an unexpected message, namely:\n"
                    + str(login_nordvpn)
                )
        else:
            print("NordVPN login check: \33[92m\N{check mark}\33[0m")

        # disconnect from VPN if necessary
        terminate = subprocess.Popen(["nordvpn", "d"], stdout=DEVNULL)

        # provide opportunity to execute additional settings.#
        settings_input_message = (
            "\n\033[34mDo you want to execute additional settings?\033[0m"
        )
        while additional_settings_needed == 1 and skip_settings == 0:
            additional_settings = input(
                settings_input_message + "\n_________________________\n\n"
                "Press enter to continue\n"
                "Type 'help' for available options\n"
            ).strip()
            if additional_settings == "help":
                options_linux = (
                    pkg_resources.open_text(NordVPN_options, "options_linux.txt")
                    .read()
                    .split("\n")
                )
                for line in options_linux:
                    print(line)
                additional_settings = input("").strip()

            additional_settings = str(additional_settings).split(" ")
            if len(additional_settings[0]) > 0:
                settings_input_message = additional_settings_linux(additional_settings)
                if any(
                    re.findall(
                        r"done|already been executed",
                        settings_input_message,
                        re.IGNORECASE,
                    )
                ):
                    additional_settings_list.append(additional_settings)
            else:
                additional_settings_needed = 0

        # however, if provided, just skip the additional settings option and execute the stored settings.#
        if "instructions" in locals():
            if len(instructions["additional_settings"][0][0]) > 0:
                print("Executing stored additional settings....\n")
                for count, instruction in enumerate(
                    instructions["additional_settings"]
                ):
                    print(
                        "Executing stored setting #"
                        + str(count + 1)
                        + ": "
                        + str(instruction)
                    )
                    additional_settings_linux(instruction)
            else:
                pass

    else:
        raise Exception(
            "I'm sorry, NordVPN switcher only works for Windows and Linux machines."
        )

    ###provide settings for VPN rotation###

    ##open available options and store these in a dict##
    areas_list = (
        pkg_resources.open_text(NordVPN_options, "countrylist.txt").read().split("\n")
    )
    country_dict = {
        "countries": areas_list[0:60],
        "europe": areas_list[0:36],
        "americas": areas_list[36:44],
        "africa east india": areas_list[49:60],
        "asia pacific": areas_list[49:60],
        "regions australia": areas_list[60:65],
        "regions canada": areas_list[65:68],
        "regions germany": areas_list[68:70],
        "regions india": areas_list[70:72],
        "regions united states": areas_list[72:87],
        "special groups": areas_list[87 : len(areas_list)],
    }

    ##provide input if needed##
    while input_needed > 0:
        if input_needed == 2:
            print("\nYou've entered a list of connection options. Checking list...\n")
            try:
                settings_servers = [area.lower() for area in area_input]
                settings_servers = ",".join(settings_servers)
            except TypeError:
                raise Exception(
                    "I expected a list here. Are you sure you've not entered a string or some other object?\n "
                )

        else:
            settings_servers = (
                input(
                    "\n\033[34mI want to connect to...\n"
                    "_________________________\n"
                    "Type 'help' for available options\n\033[0m"
                )
                .strip()
                .lower()
            )
        # define help menu#
        if settings_servers.lower().strip() == "help":
            # notation for specific servers differs between Windows and Linux.#
            if opsys == "Windows":
                notation_specific_server = " (e.g. Netherlands #742,Belgium #166)\n"
            else:
                notation_specific_server = " (e.g. nl742,be166)\n"

            settings_servers = (
                input(
                    "\nOptions:\n"
                    "##########\n"
                    "* type 'quick' to choose quickconnect \n"
                    "* type 'complete rotation' to rotate between all available NordVPN servers\n"
                    "* Single country or local region (e.g.Germany)\n"
                    "* Regions within country (e.g. regions united states')\n"
                    "* World regions (europe/americas/africa east india/asia pacific)\n"
                    "* Random multiple countries and/or local regions (e.g.France,Netherlands,Chicago)\n"
                    "* Random n countries (e.g. random countries 10)\n"
                    "* Random n countries within larger region (e.g. random countries europe 5)\n"
                    "* Random n regions in country (e.g. random regions United States 6)\n"
                    "* Specialty group name (e.g. Dedicated IP,Double VPN)\n"
                    "* Specific list of servers" + notation_specific_server
                )
                .strip()
                .lower()
            )

        # set base command according to running os#
        if opsys == "Windows":
            nordvpn_command = ["nordvpn", "-c"]
        if opsys == "Linux":
            nordvpn_command = ["nordvpn", "c"]

        # create sample of regions from input.#
        # 1. if quick connect#
        if settings_servers == "quick":
            if input_needed == 1:
                quickconnect_check = input(
                    "\nYou are choosing for the quick connect option. Are you sure? (y/n)\n"
                )
                if "y" in quickconnect_check:
                    sample_countries = [""]
                    input_needed = 0
                    pass
            if input_needed == 2:
                sample_countries = [""]
                input_needed = 0
            else:
                print("\nYou are choosing for the quick connect option.\n")
        # 2. if completely random rotation
        elif settings_servers == "complete rotation":
            print("\nFetching list of all current NordVPN servers...\n")
            for i in range(120):
                try:
                    filtered_servers = get_nordvpn_servers()
                    if opsys == "Windows":
                        nordvpn_command.append("-n")
                        sample_countries = filtered_servers["windows_names"]
                    else:
                        sample_countries = filtered_servers["linux_names"]
                except:
                    time.sleep(0.5)
                    continue
                else:
                    input_needed = 0
                    break
            else:
                raise Exception(
                    "\nI'm unable to fetch the current NordVPN serverlist. Check your internet connection.\n"
                )
        # 3. if provided specific servers. Notation differs for Windows and Linux machines, so two options are checked (first is Windows, second is Linux#
        elif (
            "#" in settings_servers
            or re.compile(r"^[a-zA-Z]+[0-9]+").search(settings_servers.split(",")[0])
            is not None
        ):
            if opsys == "Windows":
                nordvpn_command.append("-n")
            sample_countries = [area.strip() for area in settings_servers.split(",")]
            input_needed = 0
        else:
            # 3. If connecting to some specific group of servers#
            if opsys == "Windows":
                nordvpn_command.append("-g")
            # 3.1 if asked for random sample, pull a sample.#
            if "random" in settings_servers:
                # determine sample size#
                samplesize = int(re.sub("[^0-9]", "", settings_servers).strip())
                # 3.1.1 if asked for random regions within country (e.g. random regions from United States,Australia,...)#
                if "regions" in settings_servers:
                    try:
                        sample_countries = country_dict[
                            re.sub("random", "", settings_servers)
                            .rstrip("0123456789.- ")
                            .lower()
                            .strip()
                        ]
                        input_needed = 0
                    except:
                        input(
                            "\n\nThere are no specific regions available in this country, please try again.\nPress enter to continue.\n"
                        )
                        if input_needed == 2:
                            input_needed = 1
                            continue
                    if re.compile(r"[^0-9]").search(settings_servers.strip()):
                        sample_countries = random.sample(sample_countries, samplesize)
                # 3.1.2 if asked for random countries within larger region#
                elif any(
                    re.findall(
                        r"europe|americas|africa east india|asia pacific",
                        settings_servers,
                    )
                ):
                    larger_region = country_dict[
                        re.sub("random|countries", "", settings_servers)
                        .rstrip("0123456789.- ")
                        .lower()
                        .strip()
                    ]
                    sample_countries = random.sample(larger_region, samplesize)
                    input_needed = 0
                # 3.1.3 if asked for random countries globally#
                else:
                    if re.compile(r"[^0-9]").search(settings_servers.strip()):
                        sample_countries = random.sample(
                            country_dict["countries"], samplesize
                        )
                        input_needed = 0
                    else:
                        sample_countries = country_dict["countries"]
                        input_needed = 0
            # 4. If asked for specific region (e.g. europe)#
            elif settings_servers in country_dict.keys():
                sample_countries = country_dict[settings_servers]
                input_needed = 0
            # 5. If asked for specific countries or regions (e.g.netherlands)#
            else:
                # check for empty input first.#
                if settings_servers == "":
                    input(
                        "\n\nYou must provide some kind of input.\nPress enter to continue and then type 'help' to view the available options.\n"
                    )
                    if input_needed == 2:
                        input_needed = 1
                        continue
                else:
                    sample_countries = [
                        area.strip() for area in settings_servers.split(",")
                    ]  # take into account possible superfluous spaces#
                    approved_regions = 0
                    for region in sample_countries:
                        if region in [area.lower() for area in areas_list]:
                            approved_regions = approved_regions + 1
                            pass
                        else:
                            input(
                                "\n\nThe region/group "
                                + region
                                + " is not available. Please check for spelling errors.\nPress enter to continue.\n"
                            )
                            if input_needed == 2:
                                input_needed = 1
                                continue
                    if approved_regions == len(sample_countries):
                        input_needed = 0

    ##fetch current ip to prevent ip leakage when rotating VPN##
    for i in range(59):
        try:
            og_ip = get_ip()
        except ConnectionAbortedError:
            time.sleep(1)
            continue
        else:
            break
    else:
        raise Exception(
            "Can't fetch current ip, even after retrying... Check your internet connection."
        )

    ##if user does not use preloaded settings##
    if "instructions" not in locals():
        # 1.add underscore if spaces are present on Linux os#
        for number, element in enumerate(sample_countries):
            if element.count(" ") > 0 and opsys == "Linux":
                sample_countries[number] = re.sub(" ", "_", element)
        else:
            pass
        # 2.create instructions dict object#
        instructions = {
            "opsys": opsys,
            "command": nordvpn_command,
            "settings": sample_countries,
            "original_ip": og_ip,
        }
        if opsys == "Windows":
            instructions["cwd_path"] = cwd_path
        if opsys == "Linux":
            instructions["additional_settings"] = additional_settings_list
            if "credentials" in locals():
                instructions["credentials"] = credentials
        # 3.save the settings if requested into .txt file in project folder#
        if save == 1:
            print("\nSaving settings in project folder...\n")
            try:
                os.remove("settings_nordvpn.txt")
            except FileNotFoundError:
                pass
            instructions_write = json.dumps(instructions)
            f = open("settings_nordvpn.txt", "w")
            f.write(instructions_write)
            f.close()

    print("\nDone!\n")
    return instructions


##########################
# rotate and terminate VPN#
##########################
# 1.rotate
def rotate_VPN(instructions=None, google_check=0):
    if instructions is None:
        instructions = saved_settings_check()

    opsys = instructions["opsys"]
    command = instructions["command"]
    settings = instructions["settings"]
    og_ip = instructions["original_ip"]

    if opsys == "Windows":
        cwd_path = instructions["cwd_path"]

    for i in range(2):
        try:
            current_ip = new_ip = get_ip()
        except urllib.error.URLError:
            print("Can't fetch current ip. Retrying...")
            time.sleep(10)
            continue
        else:
            print("\nYour current ip-address is:", current_ip)
            break
    else:
        raise Exception(
            "Can't fetch current ip, even after retrying... Check your internet connection."
        )

    for i in range(4):
        if len(settings) > 1:
            settings_pick = list([random.choice(settings)])
        else:
            settings_pick = settings

        input = command + settings_pick

        if settings[0] == "":
            print(
                "\nConnecting you to the best possible server (quick connect option)..."
            )
        else:
            print("\n\33[34mConnecting you to", settings_pick[0], "...\n\33[0m")

        try:
            if opsys == "Windows":
                new_connection = subprocess.Popen(input, shell=True, cwd=cwd_path)
                new_connection.wait(50)
            else:
                new_connection = check_output(input)
                print(
                    "Found a server! You're now on "
                    + re.search(
                        "(?<=You are connected to )(.*)(?=\()", str(new_connection)
                    )[0].strip()
                )
        except:
            print(
                "\n An unknown error occurred while connecting to a different server! Retrying with a different server...\n"
            )
            time.sleep(15)
            continue

        for i in range(12):
            try:
                new_ip = get_ip()
            except:
                time.sleep(5)
                continue
            else:
                if new_ip in [current_ip, og_ip]:
                    time.sleep(5)
                    continue
                else:
                    break
        else:
            pass

        if new_ip in [current_ip, og_ip]:
            print("ip-address hasn't changed. Retrying...\n")
            time.sleep(10)
            continue
        else:
            print("your new ip-address is:", new_ip)

        if google_check == 1:
            print(
                "\n\33[33mPerforming captcha-check on Google search and Youtube...\n"
                "---------------------------\33[0m"
            )
            try:
                google_search_check = BeautifulSoup(
                    requests.get(
                        "https://www.google.be/search?q=why+is+python+so+hard"
                    ).content,
                    "html.parser",
                )
                youtube_video_check = BeautifulSoup(
                    requests.get("https://www.youtube.com/watch?v=dQw4w9WgXcQ").content,
                    "html.parser",
                )

                google_captcha = google_search_check.find("div", id="recaptcha")
                youtube_captcha = youtube_video_check.find("div", id="recaptcha")

                if None not in (google_captcha, youtube_captcha):
                    print("Google throws a captcha. I'll pick a different server...")
                    time.sleep(5)
                    continue
            except:
                print("Can't load Google page. I'll pick a different server...")
                time.sleep(5)
                continue
            else:
                print(
                    "Google and YouTube don't throw any Captcha's: \33[92m\N{check mark}\33[0m"
                )
                break
        else:
            break

    else:
        raise Exception(
            "Unable to connect to a new server. Please check your internet connection.\n"
        )

    print("\nDone! Enjoy your new server.\n")


# 2.terminate
def terminate_VPN(instructions=None):
    if instructions is None:
        instructions = saved_settings_check()

    opsys = instructions["opsys"]
    if opsys == "Windows":
        cwd_path = instructions["cwd_path"]

    print("\nDisconnecting...")
    if opsys == "Windows":
        terminate = subprocess.Popen(
            ["nordvpn", "-d"], shell=True, cwd=cwd_path, stdout=DEVNULL
        )
    else:
        terminate = subprocess.Popen(["nordvpn", "d"], stdout=DEVNULL)

    terminate.wait()
    print("Done!")



#########################################################################################
# Global variables
#########################################################################################

from termcolor import cprint
import urllib3

# TODO
# def integrity_check(url=MD5_CHECKSUM_URL):
#     """ Check the integrity of the application """
#     if open("{}/docs/checksum.md5".format(PATH)).read() == urllib3.urlopen(url).read():
#         pass
#     else:
#         checksum_fail = "MD5 sums did not match from origin master, "
#         checksum_fail += "integrity check has failed, this could be because "
#         checksum_fail += "there is a new version available."
#         cprint(checksum_fail)
#         update = prompt("Would you like to update to the latest version[y/N]: ")
#         if update.upper().startswith("Y"):
#             update_pybelt()
#         else:
#             pass

USER_AGENTS = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:88.0) Gecko/20100101 Firefox/88.0",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:88.0) Gecko/20100101 Firefox/88.0",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 14_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPad; CPU OS 14_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Linux; Android 10; SM-A505FN) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.152 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; Android 10; SM-G975F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.152 Mobile Safari/537.36",
    "Mozilla/5.0 (Linux; U; Android 10; en-us; Redmi Note 9 Pro) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.93 Mobile Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.182 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",
    "Mozilla/5.0 (X11; CrOS x86_64 13729.56.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.82 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.190 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.192 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.82 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.82 Safari/537.36",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPad; CPU OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Linux; Android 11; Pixel 4 XL) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.82 Mobile Safari/537.36",
    "Wget/1.15 (linux-gnu)",
    "Lynx/2.8.8pre.4 libwww-FM/2.14 SSL-MM/1.4.1 GNUTLS/2.12.23",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.157 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36",
    "Mozilla/4.0 (compatible; MSIE 9.0; Windows NT 6.1)",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)",
    "Mozilla/5.0 (Windows NT 6.1; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (Windows NT 6.2; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0)",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; Trident/7.0; rv:11.0) like Gecko",
    "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0)",
    "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/6.0)",
    "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 2.0.50727; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729)",
]



#########################################################################################
# Proxy related functions
#########################################################################################

# Round-robin proxy generator
import csv
import itertools
import json
import os
import random
import threading
import time

import requests
from termcolor import cprint
import concurrent.futures
from tqdm import tqdm
from typing import List, Literal
import sys

from utils.app_config import USER_AGENTS
from fp.fp import FreeProxy, FreeProxyException


def round_robin_proxies(proxies):
    return itertools.cycle(proxies)


# Function to check if a proxy is up
def is_proxy_alive(proxy, config, retry=0):
    """
    Check if a proxy is alive by sending a test request to Google.

    Args:
        proxy (str): The proxy to test.
        config (dict): Configuration settings.
        retry (int, optional): The number of retry attempts. Defaults to 0.

    Returns:
        tuple: A tuple containing a boolean indicating if the proxy is alive and the proxy itself.
    """
    try:
        cprint(
            f"Testing proxy {proxy}, retry n° {retry} ...",
            "yellow",
            file=sys.stderr,
        )
        # TODO use request_manager
        response = requests.get(
            "http://www.google.com/search?q=test",
            proxies={"http": proxy, "https": proxy},
            timeout=config["proxy_mean_delay"],
            headers={"User-Agent": random.choice(USER_AGENTS)},
            verify=False,
        )
        if response and response.status_code == 429 and retry < 3:
            retry_after = int(response.headers.get("Retry-After", 60))
            cprint(
                f"Proxy {proxy}: Retry after {retry_after} secs ...",
                "red",
                file=sys.stderr,
            )
            time.sleep(retry_after)
            return is_proxy_alive(proxy=proxy, retry=retry + 1, config=config)
        return response.status_code == 200, proxy
    except requests.exceptions.RequestException as e:
        cprint(
            f"Proxy {proxy} error with : {e}",
            "red",
            file=sys.stderr,
        )
        return False, proxy


# Load proxies from file
def load_proxies(file="vpn_proxies/proxies/free-proxy-list.txt"):
    with open(file, "r") as file:
        return [line.strip() for line in file if line.strip()]


def remove_unavailable_proxies(proxies: List[str], config: dict) -> List[str]:
    raise NotImplementedError("This function is not implemented yet.")


def prepare_proxies(proxy, config):
    if proxy and "username:password" in proxy:
        nord_vpn_user_pass = random.choice(config["nord_vpn_login"])
        proxy = proxy.replace("username", nord_vpn_user_pass[0]).replace(
            "password", nord_vpn_user_pass[1]
        )
        proxies = {"https": proxy}

    else:
        if not "http" in proxy or not "socks" in proxy:
            proxy = "http://" + proxy
        proxies = {"http": proxy, "https": proxy}
    return proxies


def get_proxies_and_cycle(config):
    """Get proxies and proxy cycle.

    This function retrieves the proxies from the configuration and sets up a proxy cycle.
    If the configuration specifies the use of a proxy, it checks if there are any proxies available.
    If no proxies are available and the use of a proxy is required, it prints an error message and exits.
    If the use of a proxy is not required, it sets the proxies list to contain a single None value.
    Finally, it sets up a proxy cycle using the retrieved proxies and returns both the proxies list and the proxy cycle.

    Args:
        config (dict): The configuration dictionary containing the proxy settings.

    Returns:
        tuple: A tuple containing the proxies list and the proxy cycle.
    """
    proxies = config["proxies"]
    if config["use_proxy"] and len(proxies) == 0:
        cprint(
            f"Using proxies -> you should have at least one UP",
            "red",
            file=sys.stderr,
        )
        exit()

    if not config["use_proxy"]:
        proxies = [None]

    proxy_cycle = round_robin_proxies(proxies)
    return proxies, proxy_cycle


def setup_proxies(config):
    """Set up proxies based on the provided configuration.

    Args:
        config (dict): A dictionary containing the configuration options.

    Returns:
        None
    """
    proxies = []
    if config["use_free_proxy_file"]:
        cprint("Loading proxies from file ...", "yellow", file=sys.stderr)
        proxies = load_proxies()
        proxies_cp = proxies.copy()
        dead_proxies = 0
        total_proxies = len(proxies)

        lock = threading.Lock()
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            future_to_search = {
                executor.submit(is_proxy_alive, proxy, config, 0): proxy
                for proxy in proxies_cp
            }
            for future in tqdm(
                concurrent.futures.as_completed(future_to_search),
                total=len(future_to_search),
                desc="Checking proxies",
                unit="proxy",
                leave=True,
                position=0,
            ):
                result = future.result()
                if result:
                    with lock:
                        if not result[0]:
                            dead_proxies += 1
                            cprint(
                                f"Removing dead proxy {result[1]}, dead proxies {dead_proxies}/{total_proxies}",
                                "red",
                                file=sys.stderr,
                            )
                            proxies.remove(result[1])
            cprint(f"Up free proxies: {len(proxies)}")

    if config["use_free_proxy"]:
        cprint("Using Free proxies ", "yellow", file=sys.stderr)
        try:
            free_proxy = FreeProxy(
                google=None, rand=True, https=True, timeout=10
            ).get_proxy_list(repeat=False)
            proxies += free_proxy
        except FreeProxyException as e:
            cprint(f"FreeProxyException: {e}", "red", file=sys.stderr)

    if config["use_nordvpn_proxy"]:
        config["nord_vpn_login"] = json.loads(config["nord_vpn_login"])
        cprint("Using NordVPN proxies ", "yellow", file=sys.stderr)
        if os.path.exists("vpn_proxies/proxies/nordvpn_login.csv"):
            with open("vpn_proxies/proxies/nordvpn_login.csv", "r") as file:
                nordvpn = list(csv.reader(file))
                for i in range(1, len(nordvpn)):
                    config["nord_vpn_login"].append([nordvpn[i][0], nordvpn[i][1]])
                cprint(
                    f"You have NordVPN account using these proxies {config['nord_vpn_login']}",
                    "green",
                    file=sys.stderr,
                )
            # https://stackoverflow.com/questions/64516109/how-to-use-nordvpn-servers-as-proxy-for-python-requests
            # TODO: curl -s https://nordvpn.com/api/server | jq -r ".[] | select(.features.socks==true) | [.domain, .name] | @tsv"
            with open("vpn_proxies/proxies/nordvpn-proxy-list.txt", "r") as file:
                for line in file.readlines():
                    line = line.replace("\n", "")
                    cprint(f"NordVPN Proxy: {line}", "yellow", file=sys.stderr)
                    p = (
                        "socks5h://"  # socks5h enable hostname resolution
                        + "username"
                        + ":"
                        + "password"
                        + "@"
                        + line
                        + ":1080"
                    )
                    proxies += [p]
        else:
            cprint(
                "You need to provide your NordVPN login details in proxies/nordvpn_login.csv",
                "red",
                file=sys.stderr,
            )
    config["proxies"] = list(set(proxies))
    cprint(f"Proxy: {config['proxies']}", "green", file=sys.stderr)


# TODO: https://stackoverflow.com/questions/55872164/how-to-rotate-proxies-on-a-python-requests



# Setup logging
from datetime import datetime
import sys
import threading

today_date = datetime.now()
orig_stdout = sys.stderr
f = open(f'outputs/logs/"{today_date}".log', "w")
lock_stderr = threading.Lock()


class Unbuffered:
    def __init__(self, stream):
        self.stream = stream

    def flush(self):
        pass

    def write(self, data):
        with lock_stderr:
            self.stream.write(data)
            f.write(data)  # Write the data of stdout here to a text file as well
            # self.stream.flush()
            # f.flush()


sys.stderr = Unbuffered(sys.stderr)



# -*- coding: utf-8 -*-

from urllib.parse import parse_qs
from http.cookies import SimpleCookie


def parse_post_data(data):
    """
    parses the post data
    """
    result = parse_qs(data, strict_parsing=True)
    for key in result:
        if len(result[key]) == 1:
            result[key] = result[key][0]
            data = result

    return data


def parse_cookies(cookies):
    """
    converts the raw data to dict
    """
    raw_data = cookies
    cookie = SimpleCookie()
    cookie.load(raw_data)
    cookies = {}
    for key, morsel in cookie.items():
        cookies[key] = morsel.value

    return cookies


def extract_headers(headers):
    """
    extracts the headers and returns a dict
    """
    headers = headers.replace("\\n", "\n")
    return parse_headers(headers)


def parse_headers(string):
    """
    parses the headers and creates a dict
    """
    new_headers = {}
    for line in string.split("\n"):
        if len(line) > 1:
            splitted = line.split(":")
            new_headers[splitted[0]] = ":".join(splitted[1:]).strip()

    return new_headers



from setuptools import setup, find_packages


setup(
    name='bounty-dork',
    version='0.0.1',
    url='https://github.com/ElNiak/BountyDork',
    description='Automated Google Dorking tool for pentesters without API keys.',
    long_description=open('README.md').read(),
    include_package_data=True,
    author='ElNiak',
    author_email='christophe.crochet@uclouvain.be',
    long_description_content_type='text/markdown',
    maintainer='ElNiak',
    maintainer_email='christophe.crochet@uclouvain.be',
    packages=find_packages(),
    package_data={'': ['bounty_dork/dorks/google/*.txt', 'bounty_dork/vpn_proxies/proxies/*.txt', 'bounty_dork/configs/*.txt']},
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)


#########################################################################################
# Google Dorking and searching functions
#########################################################################################

# Proxy-aware Google search function
import glob
import random
import threading
import time
import requests
from tqdm import tqdm
import sys
import re
import concurrent.futures
from termcolor import cprint

from utils.app_config import (
    USER_AGENTS,
)

from scraping.web_scraper import parse_google_search_results, render_js_and_get_text

from vpn_proxies.proxies_manager import get_proxies_and_cycle, prepare_proxies
from requester.request_manager import param_converter, start_request
from reporting.results_manager import (
    get_processed_dorks,
    save_dorking_query,
    google_dorking_results,
)

dork_id_lock = threading.Lock()


def google_search_with_proxy(
    dork_query,
    proxy,
    category,
    config,
    domain,
    processed_dorks,
    retries=1,
    advanced=False,
    dork_id=0,
):
    """Performs a Google search using a proxy.

    This function takes in various parameters to perform a Google search using a proxy.
    It generates a full query based on the provided dork query, config, and domain.
    It prepares the necessary parameters and proxies, and then performs the search using the `perform_searches` function.

    Args:
        dork_query (str): The dork query to be used for the Google search.
        proxy (str): The proxy to be used for the Google search.
        category (str): The category of the search.
        config (dict): The configuration file.
        domain (str): The domain to be searched.
        processed_dorks (list): A list of processed dorks.
        retries (int, optional): The number of retries for the search. Defaults to 1.
        advanced (bool, optional): Whether to use advanced search options. Defaults to False.
        dork_id (int, optional): The ID of the dork. Defaults to 0.

    Raises:
        Exception: If the config file is not provided.

    Returns:
        dict: The search results.
    """

    if not config:
        raise Exception("Config file should be provided")

    proxies = prepare_proxies(proxy, config)

    full_query = generate_dork_query(dork_query, config, domain)

    params = prepare_params(config)

    return perform_searches(
        full_query,
        proxies,
        category,
        params,
        retries,
        config,
        advanced,
        dork_id,
        processed_dorks,
        use_session=not (proxy == None),
    )


def prepare_params(config):
    return {
        "client": "ubuntu-sn",
        "channel": "fs",
        "num": config["total_output"],
        "hl": config["lang"],
    }


def perform_searches(
    full_query,
    proxies,
    category,
    params,
    retries,
    config,
    advanced,
    dork_id,
    processed_dorks,
    use_session,
):
    """Perform searches using Google dorking.

    This function performs searches using Google dorking technique. It takes various parameters
    to customize the search query and behavior.

    Args:
        full_query (str): The full search query.
        proxies (dict): A dictionary of proxies to be used for the search.
        category (str): The category of the search.
        params (dict): Additional parameters for the search.
        retries (int): The number of retries in case of failure.
        config (dict): Configuration settings for the search.
        advanced (bool): Flag indicating whether to use advanced search techniques.
        dork_id (int): The ID of the dork.
        processed_dorks (list): A list of processed dorks.
        use_session (bool): Flag indicating whether to use session for the search.

    Returns:
        int: The ID of the executed search.
    """

    params["q"] = full_query
    dork_id = execute_search_with_retries(
        full_query,
        proxies,
        category,
        params,
        retries,
        config,
        advanced,
        dork_id,
        processed_dorks,
        use_session=use_session,
    )

    return dork_id


def execute_search_with_retries(
    query,
    proxies,
    category,
    params,
    retries,
    config,
    advanced,
    dork_id,
    processed_dorks,
    use_session=False,
):
    """
    Execute a search with retries using Google dorking.

    Args:
        query (str): The search query.
        proxies (dict): The proxies to be used for the request.
        category (str): The category of the search.
        params (dict): The parameters for the request.
        retries (int): The number of retries.
        config (dict): The configuration settings.
        advanced (bool): Whether to use advanced search techniques.
        dork_id (int): The ID of the dork.
        processed_dorks (list): The list of already processed dorks.
        use_session (bool, optional): Whether to use a session for the request. Defaults to False.

    Returns:
        int: The updated dork ID.
    """
    base_url = "https://www.google.com/search"
    headers = {
        "User-Agent": random.choice(USER_AGENTS),
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip,deflate",
        "Connection": "close",
        "DNT": "1",
        "cache-control": "max-age=0",
        "Upgrade-Insecure-Requests": "1",
    }

    if query in processed_dorks:
        cprint(
            f"Skipping already processed dork: {query}",
            "yellow",
            file=sys.stderr,
        )
        return dork_id

    for retry_no in range(retries):
        if use_session:
            cprint(
                f"Searching for GET - Session (n° {retry_no}): {base_url} \n\t - parameters {params} \n\t - headers {headers} \n\t - {category} - with proxy {proxies} ...",
                "yellow",
                file=sys.stderr,
            )
            with requests.Session() as session:
                response = start_request(
                    config=config,
                    proxies=proxies,
                    base_url=base_url,
                    GET=True,
                    headers=headers,
                    params=params,
                    is_json=False,
                    secured=(
                        True
                        if proxies
                        and "https" in proxies
                        and proxies["https"]
                        and "socks" in proxies["https"]
                        else False
                    ),
                    session=session,
                    cookies={
                        "CONSENT": "PENDING+987",
                        "SOCS": "CAESHAgBEhJnd3NfMjAyMzA4MTAtMF9SQzIaAmRlIAEaBgiAo_CmBg",
                    },
                )
        else:
            cprint(
                f"Searching for GET (n° {retry_no}): {base_url} \n\t - parameters {params} \n\t - headers {headers} \n\t - {category} - with proxy {proxies} ...",
                "yellow",
                file=sys.stderr,
            )
            response = start_request(
                config=config,
                proxies=proxies,
                base_url=base_url,
                GET=True,
                headers=headers,
                params=params,
                is_json=False,
                secured=(
                    True
                    if proxies
                    and "https" in proxies
                    and proxies["https"]
                    and "socks" in proxies["https"]
                    else False
                ),
                cookies={
                    "CONSENT": "PENDING+987",
                    "SOCS": "CAESHAgBEhJnd3NfMjAyMzA4MTAtMF9SQzIaAmRlIAEaBgiAo_CmBg",
                },
            )

        urls = []
        if response:
            urls = parse_google_search_results(proxies, advanced, query, response.text)
            if (not urls or len(urls) == 0) and config["use_selenium"]:
                cprint(
                    f"Parsing for google search failed for {query} - retrying with selenium...",
                    "red",
                    file=sys.stderr,
                )
                html_content = render_js_and_get_text(
                    param_converter(params, url=base_url)
                )
                urls = parse_google_search_results(
                    proxies, advanced, query, html_content
                )
        result = dork_id, category, urls, query
        # save_dorking_query(result, config)
        with dork_id_lock:
            google_dorking_results.append((result, config))
        # with dork_id_lock:
        #     dork_id += 1
    # TODO to be faster also record non functionnal dork
    return dork_id


google_dork_tags = [
    "site:",  # Search within a specific site or domain
    "intitle:",  # Search for pages with a specific word in the title
    "inurl:",  # Search for pages with a specific word in the URL
    "intext:",  # Search for pages with a specific word in the text
    "filetype:",  # Search for files of a specific type (e.g., PDF, DOC)
    "ext:",  # Similar to filetype:, search for files with a specific extension
    "link:",  # Search for pages that link to a specific URL
    "cache:",  # Display Google's cached version of a page
    "related:",  # Find sites related to a given URL
    "info:",  # Show information about a specific URL
    "define:",  # Provide definitions for words
    "insubject:",  # Search for pages with a specific word in the subject (used in Google Groups)
    "daterange:",  # Search within a specific date range (using Julian dates)
    "allintitle:",  # Search for pages with all specified words in the title
    "allinurl:",  # Search for pages with all specified words in the URL
    "allintext:",  # Search for pages with all specified words in the text
    "allinanchor:",  # Search for pages with all specified words in the anchor text
    "author:",  # Search for articles written by a specific author (used in Google Groups)
    "group:",  # Search within a specific Google Group (used in Google Groups)
    "numrange:",  # Search within a specific number range
    "movie:",  # Search for information about a specific movie
    "map:",  # Search for maps
    "weather:",  # Get weather information
    "stocks:",  # Get stock information
    "phonebook:",  # Search for phonebook listings
    "book:",  # Search for books
    "safesearch:",  # Enable or disable safe search
    "blogurl:",  # Search for blogs at a specific URL
    "location:",  # Search for pages related to a specific location
    "before:",  # Search for pages before a specific date
    "after:",  # Search for pages after a specific date
]


def generate_dork_query(query, config, domain):
    """Generate a dork query for Google dorking.

    This function takes a query, configuration, and domain as input and generates a dork query
    for Google dorking. It cleans up the query by removing existing inurl: and intext: tags,
    ensures the query is properly enclosed in quotes if it contains quotes, and incorporates
    the subdomain into the search query if specified. It also appends the file extension to the
    query if specified in the configuration.

    Args:
        query (str): The query string.
        config (dict): The configuration settings.
        domain (str): The domain to incorporate into the search query.

    Returns:
        str: The generated dork query.

    """
    # Clean up the query by removing existing inurl: and intext: tags
    if len(query) > 0:
        for tag in ["inurl:", "intext:"]:
            query = query.replace(tag, "")

        # Ensure the query is properly enclosed in quotes if it contains quotes
        # if '"' in query:
        if not query.startswith('"'):
            query = '"' + query
        if not query.endswith('"'):
            query = query + '"'

        in_url_query = "inurl:" + query
        in_text_query = "intext:" + query

        query = in_url_query + " | " + in_text_query

        query = query  # + " | "  # + "inurl:&"

    # Incorporate subdomain into the search query if specified
    if domain:
        # Remove any existing site: tag and its value
        full_query = []
        query = re.sub(r"site:[^\s]+", "", query)
        to_search = f"site:{domain}"
        if len(query) > 0:
            full_query = f"({to_search}) & ({query})".strip()
        else:
            full_query = f"({to_search})".strip()
    else:
        full_query = f"({query})".strip()

    if config["extension"] and len(config["extension"]) > 0:
        full_query = full_query + f" & filetype:{config['extension']}"

    return full_query  # Indicate failure after retries


def filter_search_tasks(search_tasks, processed_dorks):
    """
    Filters out the already processed dorks from search tasks.

    Args:
        search_tasks (dict): A dictionary containing search tasks categorized by their respective categories.
        processed_dorks (list): A list of already processed dorks.

    Returns:
        dict: A dictionary containing filtered search tasks with the already processed dorks removed.

    """
    filtered_tasks = {}
    for category, dorks in search_tasks.items():
        filtered_tasks[category] = [
            dork for dork in dorks if dork not in processed_dorks
        ]
    return filtered_tasks


def launch_google_dorks_and_search_attack(config, categories):
    """Launches a Google dorks and search attack.

    This function takes a configuration object and a list of categories as input.
    It performs a Google dorks search for each category and launches a search attack
    using the obtained dorks.

    Args:
        config (dict): A configuration object containing various settings for the attack.
        categories (list): A list of categories to perform the search attack on.

    Raises:
        NotImplementedError: If VPN is enabled in the configuration, as VPN is not supported in this version.
        Exception: If any other error occurs during the search attack.

    Returns:
        None
    """
    start_time = time.time()
    try:

        proxies, proxy_cycle = get_proxies_and_cycle(config)

        search_tasks = {}

        for category in categories:
            search_tasks[category] = []
            dork_files = glob.glob(
                f"dorks/google/{category}/*.txt", recursive=True
            )
            for dork_file in dork_files:
                with open(dork_file, "r") as file:
                    lines = file.readlines()
                    dorks = [line.strip() for line in lines]
                search_tasks[category] += dorks

        cprint(
            f"Total number of dorks: {sum([len(search_tasks[task]) for task in search_tasks])}",
            "yellow",
            file=sys.stderr,
        )
        processed_dorks = get_processed_dorks(config)

        if not search_tasks:
            cprint(f"No dorks to process.", "red", file=sys.stderr)
            return

        if config["use_vpn"]:
            raise NotImplementedError(
                "VPN is not supported in this version - Error in library"
            )
            thread = threading.Thread(target=change_vpn)
            thread.start()

        number_of_worker = 30 # min(len(proxies)*2, 30)
        cprint(f"Number of workers: {number_of_worker}", "yellow", file=sys.stderr)

        search_tasks_with_proxy = []
        for task in search_tasks:
            for domain in config["subdomain"]:
                for dork in search_tasks[task]:
                    proxy = next(proxy_cycle)
                    search_tasks_with_proxy.append(
                        {
                            "dork": dork,
                            "proxy": proxy,
                            "category": task,
                            "domain": domain,
                        }
                    )
        cprint(
            f"Total number of dorks: {len(search_tasks_with_proxy)}",
            "yellow",
            file=sys.stderr,
        )
        # TODO https://stackoverflow.com/questions/65832061/threadpoolexecutor-keyboardinterrupt
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=number_of_worker
        ) as executor:
            future_to_search = {
                executor.submit(
                    google_search_with_proxy,
                    task["dork"],
                    task["proxy"],
                    task["category"],
                    config,
                    task["domain"],
                    processed_dorks,
                ): task
                for task in search_tasks_with_proxy
            }
            try:
                for future in tqdm(
                    concurrent.futures.as_completed(future_to_search),
                    total=len(future_to_search),
                    desc="Searching for vulnerable website",
                    unit="site",
                ):
                    future.result()
            except KeyboardInterrupt:
                cprint(
                    "Process interrupted by user during google dorking phase ... Saving results",
                    "red",
                    file=sys.stderr,
                )

                end_time = time.time()
                cprint(
                    "Total time taken: " + str(end_time - start_time), "green", file=sys.stderr
                )
                executor._threads.clear()
                concurrent.futures.thread._threads_queues.clear()
                # https://stackoverflow.com/questions/49992329/the-workers-in-threadpoolexecutor-is-not-really-daemon
                for result, config in google_dorking_results:
                    save_dorking_query(result, config)
                exit()

        end_time = time.time()
        cprint(
            "Total time taken: " + str(end_time - start_time), "green", file=sys.stderr
        )

        cprint(
            f"Saving dorks - Total number of dorks processed: {len(google_dorking_results)}",
            "green",
            file=sys.stderr,
        )

        # TODO remove duplicate url and merge dorks
        for result, config in google_dorking_results:
            save_dorking_query(result, config)
    except Exception as e:
        cprint(f"Error searching for dorks: {e}", "red", file=sys.stderr)
        raise e



from vpn_proxies.nordvpn_switcher import initialize_VPN, rotate_VPN
import time

##############
## WINDOWS ###
##############

# [1] save settings file as a variable

instructions = (
    initialize_VPN()
)  # this will guide you through a step-by-step guide, including a help-menu with connection options

for i in range(3):
    rotate_VPN(instructions)  # refer to the instructions variable here
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)

# [2] if you'd like to skip the step-by-step menu (because you want to automate your script fully without any required human intervention, use the area_input parameter

instructions = initialize_VPN(
    area_input=["Belgium,France,Netherlands"]
)  # <-- Be aware: the area_input parameter expects a list, not a string

for i in range(3):
    rotate_VPN(instructions)  # refer to the instructions variable here
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)

# [3] of course, you can try one of the built-in randomizers if you can't be bothered with selecting specific regions

# The following options are avilable:
# random countries X
# random countries europe X
# random countries americas X
# random countries africa east india X
# random countries asia pacific X
# random regions australia X
# random regions canada X
# random regions germany X
# random regions india X
# random regions united states X

instructions = initialize_VPN(area_input=["random countries europe 8"])

for i in range(3):
    rotate_VPN(instructions)  # refer to the instructions variable here
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)

# [4] instead of saving the instructions as a variable, you could save your settings in your work directory. Just set the save parameter to 1

initialize_VPN(save=1)

for i in range(3):
    rotate_VPN()  # Call the rotate_VPN without any parameter. It will look for a settings file in your work directory
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)

# [5] If you'd like to use an already saved settings file in your work directory (for example: a colleague/friend has sent you his/her settings file),
# use the stored settings parameter

initialize_VPN(
    save=1, area_input=["random countries 20"]
)  # save settings file to work directory
print(
    "Imagine you close your python environment and run your script on a later date. Just load your saved settings by running the following line of code:\n"
)
time.sleep(7)
initialize_VPN(
    stored_settings=1
)  # the function will look for a settingsfile in your work directory, launch NordVPN, disconnect if necessary and validate the stored settings file.

for i in range(3):
    rotate_VPN()
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)


# [6] save settings file to work directory and perform a 'complete rotation'

# --> a complete rotation will fetch a list of the currently 4000+ active available servers of NordVPN.
# The rotation will rotate between all the available servers completely by random.
# The difference with picking particular regions is that NordVPN will NOT pick the 'most appropriate' (fastest) server within a particular region when rotating.
# Instead, complete rotation will pick a specific server at random, which mens you're unlikely to revisit the same server twice in a row.
# Because of this, the 'complete rotation' option is ideal for webscraping purposes

initialize_VPN(save=1, area_input=["complete rotation"])

for i in range(3):
    rotate_VPN()
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)

# [7] You can be as creative as you like. For example, the following code will perform an infinite loop of picking a random server every hour

instructions = initialize_VPN(area_input=["complete rotation"])

while True:
    rotate_VPN(instructions)
    time.sleep(3600)

# [8] To implement the google and youtube captcha-check, use the google_check parameter

instructions = initialize_VPN(area_input=["random regions united states 8"])

for i in range(3):
    rotate_VPN(google_check=1)
    print("\nDo whatever you want here (e.g.scraping). Pausing for 10 seconds...\n")
    time.sleep(10)

##############
## LINUX #####
##############

# [1] Perform a complete rotation and skip the settings menu for complete automation
# the 'skip settings' parameter is only available for Linux users (since setting additional settings such as whitelisting ports is only available on Linux)

instr = initialize_VPN(area_input=["complete rotation"], skip_settings=1)

for i in range(3):
    rotate_VPN(instr)
    print("\nDo whatever you want here (e.g. scraping). Pausing for 10 seconds...\n")
    time.sleep(10)
